/*
 * Copyright Contributors to the fusa-gcc-plugin project
 *
 * SPDX-License-Identifier: LGPL-2.1-or-later
 */
#include "process.hpp"
#include "cryptoutils.hpp"
#include "json.hpp"

using namespace std;

int svector_to_json_array(const vector<string> &v, json_object **out) {
  _json_object MAKE_PTR(object, json_object_new_array());
  int r = 0;
  *out = NULL;

  if (v.empty()) {
    *out = json_object_new_null();
    return 0;
  }

  for (const string &s : v) {
    json_object *cs = json_object_new_string(s.c_str());
    if (cs == NULL)
      return 1;

    r = json_object_array_add(object.get(), cs);
    if (r != 0)
      return r;
  }

  *out = object.release();

  return 0;
}

int serialize_candidate(const candidate &candidate, json_object **out) {
  _json_object MAKE_PTR(object, json_object_new_object());
  json_object *o = NULL;

  int r = 0;
  *out = NULL;

  json_object *name = json_object_new_string(candidate.name.c_str());
  if (name == NULL)
    return 1;

  r = json_object_object_add(object.get(), "name", name);
  if (r != 0)
    return r;

  json_object *file =
      json_object_new_string(candidate.decl_source_file.c_str());
  if (file == NULL)
    return 1;

  r = json_object_object_add(object.get(), "file", file);
  if (r != 0)
    return r;

  json_object *file_sha256 = json_object_new_string(
      crypto::sha256_file(candidate.decl_source_file.c_str()).c_str());
  if (file == NULL)
    return 1;

  r = json_object_object_add(object.get(), "file-sha256", file_sha256);
  if (r != 0)
    return r;

  r = svector_to_json_array(candidate.class_hierarchy, &o);
  if (r != 0)
    return r;

  _json_object MAKE_PTR(class_hierarchy, o);
  r = json_object_object_add(object.get(), "class_hierarchy",
                             json_object_get(class_hierarchy.get()));
  if (r != 0)
    return r;

  r = svector_to_json_array(candidate.namespace_hierarchy, &o);
  if (r != 0)
    return r;

  _json_object MAKE_PTR(namespace_hierarchy, o);
  r = json_object_object_add(object.get(), "namespace_hierarchy",
                             json_object_get(namespace_hierarchy.get()));
  if (r != 0)
    return r;

  *out = object.release();

  return 0;
}

int serialize_output(const candidate::set &apis, string &out) {
  int r = 0;

  _json_object MAKE_PTR(root, json_object_new_object());
  _json_object MAKE_PTR(json_symbols, json_object_new_array());

  for (const candidate &c : apis) {
    json_object *json_candidate;

    r = serialize_candidate(c, &json_candidate);
    if (r != 0)
      return r;

    r = json_object_array_add(json_symbols.get(), json_candidate);
    if (r != 0) {
      json_object_put(json_candidate);
      return r;
    }
  }

  r = json_object_object_add(root.get(), "symbols", json_symbols.release());
  if (r != 0)
    return r;

  _json_object MAKE_PTR(json_version, json_object_new_int(expected_version));
  r = json_object_object_add(root.get(), "version", json_version.release());
  if (r != 0)
    return r;

  /*
  TODO: This should be captured externally for the entire assembly
  of all outputs. Running `dnf update` in the background while this process
  is ongoing is simply not supported.

  auto name = process::run_command(
      {"/usr/bin/bash", "-c", "source /etc/os-release ; echo $PRETTY_NAME"});

  json_object *json_name = json_object_new_string(name.stdout[0].c_str());
  if (json_name == NULL) {
    return 1;
  }

  r = json_object_object_add(root.get(), "name", json_name);
  if (r != 0) return r;

  auto allrpms = process::run_command({"/usr/bin/rpm", "-qa"});

  json_object *json_rpms;
  r = svector_to_json_array(allrpms.stdout, &json_rpms);
  if (r != 0) return r;

  r = json_object_object_add(root.get(), "rpms", json_rpms);
  if (r != 0) {
    json_object_put(json_rpms);
    return r;
  }*/

  out = json_object_to_json_string_ext(
      root.release(), JSON_C_TO_STRING_PRETTY | JSON_C_TO_STRING_NOSLASHESCAPE);

  return 0;
}

int deserialize_symbol(json_object *symbol, certified_symbol &out) {
  json_object *p = NULL;
  certified_symbol tmp;

  if (!json_object_object_get_ex(symbol, "symbol", &p)) {
    //cerr << "Missing name" << endl;
    return 1;
  }

  tmp.name = json_object_get_string(p);

  if (!json_object_object_get_ex(symbol, "file", &p)) {
    //cerr << "Missing file" << endl;
    return 1;
  }

  tmp.file = json_object_get_string(p);

  if (!json_object_object_get_ex(symbol, "file-sha256", &p)) {
    //cerr << "Missing file-sha256" << endl;
    return 1;
  }

  tmp.file_sha256 = json_object_get_string(p);

  if (!json_object_object_get_ex(symbol, "profiles", &p)) {
    //cerr << "Missing profiles" << endl;
    return 1;
  }

  vector<string> profiles;
  size_t num_profiles = json_object_array_length(p);

  for (size_t i = 0; i < num_profiles; ++i) {
    json_object *prof = json_object_array_get_idx(p, i);
    string profile = json_object_get_string(prof);
    for (auto &c : profile)
      c = TOUPPER(c);
    profiles.push_back(profile);
  }

  copy(profiles.begin(), profiles.end(),
       inserter(tmp.profiles, tmp.profiles.end()));

  out = tmp;

  return 0;
}

int deserialize_symbols(json_object *root, vector<certified_symbol> &symbols) {
  json_object *symbols_array = NULL;
  int r = 0;

  if (!json_object_object_get_ex(root, "certified_symbols", &symbols_array)) {
    //cerr << "Configuration: Missing key 'certified_symbols'" << endl;
    return 1;
  }

  vector<certified_symbol> tmp;
  size_t num_symbols = json_object_array_length(symbols_array);

  for (size_t i = 0; i < num_symbols; ++i) {
    json_object *symbol = json_object_array_get_idx(symbols_array, i);

    certified_symbol s;
    r = deserialize_symbol(symbol, s);
    if (r != 0)
      return r;

    tmp.push_back(s);
  }

  symbols = tmp;

  return 0;
}

int load_symbols_from_file(string &file, vector<certified_symbol> &symbols) {
  int r = 0;
  vector<certified_symbol> tmp;

  json_object *root = json_object_from_file(file.c_str());
  const char *err = json_util_get_last_err();
  if (err != NULL) {
    //cerr << err << endl;
    return 1;
  }

  json_object *version_obj;
  if (!json_object_object_get_ex(root, "version", &version_obj)) {
    //cerr << "Configuration file missing version information: " << file << endl;
    return 1;
  }

  int version = json_object_get_int(version_obj);
  if (version != expected_version) {
    //cerr << "Configuration file has invalid version: " << version
    //     << " expected " << expected_version << endl;
    return 1;
  }

  r = deserialize_symbols(root, tmp);
  if (r != 0)
    return r;

  symbols = tmp;

  return 0;
}
